/*
 * FileOps.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using System.Collections;
using System.Diagnostics;
using System.IO;
using System.Reflection;

#if !MONO
using System.Security.AccessControl;
using System.Security.Principal;
#endif

using Eagle._Attributes;
using Eagle._Components.Public;
using Eagle._Components.Public.Delegates;
using Eagle._Containers.Public;

namespace Eagle._Components.Private
{
    [ObjectId("814d0e8c-c65a-4c7c-8c2e-2e6cee551509")]
    internal static class FileOps
    {
        //
        // NOTE: The EXE header signature.
        //
        private const ushort IMAGE_DOS_SIGNATURE = 0x5A4D; // "MZ"

        //
        // NOTE: The PE header signature.
        //
        private const uint IMAGE_NT_SIGNATURE = 0x00004550; // "PE\0\0"

        //
        // NOTE: This "magic" value means that we have no idea what
        //       the value for the file (or operating system) is.
        //
        internal const ushort IMAGE_NT_OPTIONAL_BAD_MAGIC = 0x0;

        //
        // NOTE: The "magic" values from the PE header for 32-bit
        //       and 64-bit executables.
        //
        private const ushort IMAGE_NT_OPTIONAL_HDR32_MAGIC = 0x010B;
        private const ushort IMAGE_NT_OPTIONAL_HDR64_MAGIC = 0x020B;

        //
        // NOTE: The offset into the file where the offset into
        //       the file of the PE header is located.
        //
        private const int filePeSignatureOffsetOffset = 0x3C;

        //
        // NOTE: The offset for the TimeDateStamp field in the
        //       IMAGE_FILE_HEADER structure from the start of the
        //       PE signature.
        //
        private const int peTimeStampOffset = 0x8;

        //
        // NOTE: The offset for the Magic field in the
        //       IMAGE_OPTIONAL_HEADER structure from the start of
        //       the PE signature.
        //
        private const int peMagicOffset = 0x18;

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if NATIVE && WINDOWS
        //
        // NOTE: The offset for the SizeOfStackReserve field in the
        //       IMAGE_OPTIONAL_HEADER structure from the start of
        //       the PE signature.  This value just happens to be
        //       the same for 32-bit and 64-bit executables.
        //
        private const int peReserveOffset = 0x60;

        ///////////////////////////////////////////////////////////////////////////////////////////////

        //
        // NOTE: The offset for the SizeOfStackCommit field in the
        //       IMAGE_OPTIONAL_HEADER structure from the start of
        //       the PE signature.  This value is different for
        //       32-bit and 64-bit executables.
        //
        private const int peCommitOffset32Bit = 0x64;
        private const int peCommitOffset64Bit = 0x68;
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        //
        // NOTE: These strings are only used by GetPeFileMagicName.
        //
        private const string magicPe32 = "PE32";
        private const string magicPe32Plus = "PE32+";

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if NATIVE && WINDOWS
        internal static readonly NativeStack.StackSize PeFileStackSize =
            GetPeFileStackSize(GlobalState.GetEntryAssembly());
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static string GetPeFileMagicName(
            ushort magic
            )
        {
            switch (magic)
            {
                case IMAGE_NT_OPTIONAL_HDR32_MAGIC:
                    return magicPe32; // (i.e. 32-bit)
                case IMAGE_NT_OPTIONAL_HDR64_MAGIC:
                    return magicPe32Plus; // (i.e. 64-bit)
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if NATIVE && TCL
        public static ushort GetPeFileMagicForOperatingSystem()
        {
            if (PlatformOps.IsWindowsOperatingSystem())
            {
                if (PlatformOps.Is32BitOperatingSystem())
                    return IMAGE_NT_OPTIONAL_HDR32_MAGIC;
                else if (PlatformOps.Is64BitOperatingSystem())
                    return IMAGE_NT_OPTIONAL_HDR64_MAGIC;
            }

            return IMAGE_NT_OPTIONAL_BAD_MAGIC;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static bool CheckPeFileArchitecture(
            string fileName,
            ref ushort magic,
            ref Result error
            )
        {
            if (PlatformOps.IsWindowsOperatingSystem())
            {
                if (GetPeFileMagic(fileName, ref magic, ref error))
                {
                    ushort osMagic = GetPeFileMagicForOperatingSystem();

                    if (osMagic != IMAGE_NT_OPTIONAL_BAD_MAGIC)
                    {
                        if (magic == osMagic)
                            return true;

                        error = String.Format(
                            "file \"{0}\" is not for this architecture " +
                            "(magic mismatch, got 0x{1:X}, wanted 0x{2:X}).",
                            fileName, magic, osMagic);
                    }
                    else
                    {
                        error = "operating system is neither 32-bit Windows nor 64-bit Windows";
                    }
                }

                return false;
            }
            else
            {
                //
                // NOTE: Not Windows, we do not even know if this operating system
                //       uses PE files for executables.
                //
                return true;
            }
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static bool GetPeFileMagic(
            string fileName,
            ref ushort magic,
            ref Result error
            )
        {
            bool result = false;

            try
            {
                //
                // NOTE: The BinaryReader.Close() method is documented to
                //       close the contained stream as well.
                //
                using (BinaryReader binaryReader = new BinaryReader(new FileStream(
                    fileName, FileMode.Open, FileAccess.Read)))
                {
                    ushort dosSignature = binaryReader.ReadUInt16();

                    if (dosSignature == IMAGE_DOS_SIGNATURE)
                    {
                        binaryReader.BaseStream.Seek(filePeSignatureOffsetOffset, SeekOrigin.Begin);

                        uint offset = binaryReader.ReadUInt32();

                        binaryReader.BaseStream.Seek(offset, SeekOrigin.Begin);

                        uint ntSignature = binaryReader.ReadUInt32();

                        if (ntSignature == IMAGE_NT_SIGNATURE)
                        {
                            binaryReader.BaseStream.Seek(
                                offset + peMagicOffset,
                                SeekOrigin.Begin);

                            magic = binaryReader.ReadUInt16(); /* NOT VALIDATED */

                            result = true;
                        }
                        else
                        {
                            error = String.Format(
                                "NT signature mismatch, got 0x{0:X}, wanted 0x{1:X}.",
                                ntSignature, IMAGE_NT_SIGNATURE);
                        }
                    }
                    else
                    {
                        error = String.Format(
                            "DOS signature mismatch, got 0x{0:X}, wanted 0x{1:X}.",
                            dosSignature, IMAGE_DOS_SIGNATURE);
                    }
                }
            }
            catch (Exception e)
            {
                error = e;
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if NATIVE && WINDOWS
        public static bool GetPeFileStackReserveAndCommit(
            string fileName,
            ref UIntPtr reserve,
            ref UIntPtr commit
            )
        {
            bool result = false;

            try
            {
                UIntPtr localReserve = UIntPtr.Zero;
                UIntPtr localCommit = UIntPtr.Zero;

                //
                // NOTE: The BinaryReader.Close() method is documented to
                //       close the contained stream as well.
                //
                using (BinaryReader binaryReader = new BinaryReader(new FileStream(
                    fileName, FileMode.Open, FileAccess.Read)))
                {
                    if (binaryReader.ReadUInt16() == IMAGE_DOS_SIGNATURE)
                    {
                        binaryReader.BaseStream.Seek(
                            filePeSignatureOffsetOffset,
                            SeekOrigin.Begin);

                        uint offset = binaryReader.ReadUInt32();

                        binaryReader.BaseStream.Seek(offset, SeekOrigin.Begin);

                        if (binaryReader.ReadUInt32() == IMAGE_NT_SIGNATURE)
                        {
                            binaryReader.BaseStream.Seek(offset + peMagicOffset,
                                SeekOrigin.Begin);

                            switch (binaryReader.ReadUInt16())
                            {
                                case IMAGE_NT_OPTIONAL_HDR32_MAGIC:
                                    {
                                        binaryReader.BaseStream.Seek(
                                            offset + peReserveOffset,
                                            SeekOrigin.Begin);

                                        localReserve = new UIntPtr(
                                            binaryReader.ReadUInt32());

                                        binaryReader.BaseStream.Seek(
                                            offset + peCommitOffset32Bit,
                                            SeekOrigin.Begin);

                                        localCommit = new UIntPtr(
                                            binaryReader.ReadUInt32());

                                        result = true;
                                        break;
                                    }
                                case IMAGE_NT_OPTIONAL_HDR64_MAGIC:
                                    {
                                        binaryReader.BaseStream.Seek(
                                            offset + peReserveOffset,
                                            SeekOrigin.Begin);

                                        localReserve = new UIntPtr(
                                            binaryReader.ReadUInt64());

                                        binaryReader.BaseStream.Seek(
                                            offset + peCommitOffset64Bit,
                                            SeekOrigin.Begin);

                                        localCommit = new UIntPtr(
                                            binaryReader.ReadUInt64());

                                        result = true;
                                        break;
                                    }
                            }
                        }
                    }
                }

                //
                // NOTE: Did we succeed in reading the
                //       value(s) from the file?
                //
                if (result)
                {
                    reserve = localReserve;
                    commit = localCommit;
                }
            }
            catch
            {
                // do nothing.
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static NativeStack.StackSize GetPeFileStackSize(
            Assembly assembly
            ) /* THREAD-SAFE */
        {
            NativeStack.StackSize stackSize = null;

            if (assembly != null)
            {
                stackSize = new NativeStack.StackSize();

                /* IGNORED */
                GetPeFileStackReserveAndCommit(assembly.Location,
                    ref stackSize.reserve, ref stackSize.commit);
            }

            return stackSize;
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static bool GetPeFileTimeStamp(
            string fileName,
            ref uint timeStamp
            )
        {
            bool result = false;

            try
            {
                uint localTimeStamp = 0;

                //
                // NOTE: The BinaryReader.Close() method is documented to
                //       close the contained stream as well.
                //
                using (BinaryReader binaryReader = new BinaryReader(new FileStream(
                    fileName, FileMode.Open, FileAccess.Read)))
                {
                    if (binaryReader.ReadUInt16() == IMAGE_DOS_SIGNATURE)
                    {
                        binaryReader.BaseStream.Seek(filePeSignatureOffsetOffset,
                            SeekOrigin.Begin);

                        uint offset = binaryReader.ReadUInt32();

                        binaryReader.BaseStream.Seek(offset, SeekOrigin.Begin);

                        if (binaryReader.ReadUInt32() == IMAGE_NT_SIGNATURE)
                        {
                            binaryReader.BaseStream.Seek(offset + peMagicOffset,
                                SeekOrigin.Begin);

                            switch (binaryReader.ReadUInt16())
                            {
                                case IMAGE_NT_OPTIONAL_HDR32_MAGIC:
                                case IMAGE_NT_OPTIONAL_HDR64_MAGIC:
                                    {
                                        binaryReader.BaseStream.Seek(
                                            offset + peTimeStampOffset,
                                            SeekOrigin.Begin);

                                        localTimeStamp =
                                            binaryReader.ReadUInt32();

                                        result = true;
                                        break;
                                    }
                            }
                        }
                    }
                }

                //
                // NOTE: Did we succeed in reading the
                //       value(s) from the file?
                //
                if (result)
                    timeStamp = localTimeStamp;
            }
            catch
            {
                // do nothing.
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode GetFileVersion(
            string fileName,
            ref FileVersionInfo version,
            ref Result error
            )
        {
            if (File.Exists(fileName))
            {
                try
                {
                    version = FileVersionInfo.GetVersionInfo(fileName);

                    return ReturnCode.Ok;
                }
                catch (Exception e)
                {
                    error = e;
                    return ReturnCode.Error;
                }
            }
            else
            {
                error = String.Format(
                    "could not read \"{0}\": no such file",
                    fileName);

                return ReturnCode.Error;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if !MONO
        private static bool HasRights(
            FileSystemRights rights,
            FileSystemRights hasRights,
            bool all
            )
        {
            if (all)
                return ((rights & hasRights) == hasRights);
            else
                return ((rights & hasRights) != 0);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static FileSystemRights FilePermissionsToFileSystemRights(
            FilePermission permissions
            )
        {
            FileSystemRights rights = (FileSystemRights)0; /* NONE */

            if (FlagOps.HasFlags(permissions, FilePermission.Read, true))
                rights |= FileSystemRights.Read;

            if (FlagOps.HasFlags(permissions, FilePermission.Write, true))
                rights |= FileSystemRights.Write;

            if (FlagOps.HasFlags(permissions, FilePermission.Execute, true))
                rights |= FileSystemRights.ExecuteFile | FileSystemRights.Traverse;

            return rights;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode IsOwner(
            string path,
            ref bool ownerStatus,
            ref Result error
            )
        {
            try
            {
                //
                // NOTE: These must be reset prior to doing anything else in case
                //       further actions fail (by raising an exception).
                //
                ownerStatus = false;

                //
                // NOTE: If the file or directory does not exist, proceed no further.
                //
                if (PathOps.PathExists(path))
                {
                    //
                    // NOTE: Get the identity of the user for the current thread.
                    //
                    WindowsIdentity identity = WindowsIdentity.GetCurrent();

                    //
                    // NOTE: Attempt to get the file security object for this file or
                    //       directory.
                    //
                    FileSystemSecurity security;

                    //
                    // NOTE: Use the correct method based on whether this path represents
                    //       a file or directory.
                    //
                    if (Directory.Exists(path))
                        security = Directory.GetAccessControl(path);
                    else
                        security = File.GetAccessControl(path);

                    //
                    // NOTE: Attempt to get the owning user and group for this file or
                    //       directory.
                    //
                    IdentityReference owner =
                        security.GetOwner(typeof(SecurityIdentifier));

                    //
                    // NOTE: If the current user is the owner of the file OR if the current
                    //       user belongs to the group that owns the file then they are the
                    //       "owner".
                    //
                    if (identity.User.Equals(owner) || identity.Groups.Contains(owner))
                        ownerStatus = true;
                }

                return ReturnCode.Ok;
            }
            catch (Exception e)
            {
                error = e;
                return ReturnCode.Error;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode AccessCheck(
            string path,
            FileSystemRights desiredRights,
            ref FileSystemRights grantedRights,
            ref bool accessStatus,
            ref Result error
            )
        {
            try
            {
                //
                // NOTE: These must be reset prior to doing anything else in case
                //       further actions fail (by raising an exception).
                //
                grantedRights = (FileSystemRights)0; /* None */
                accessStatus = false;

                //
                // NOTE: If the file or directory does not exist, proceed no further.
                //
                if (PathOps.PathExists(path))
                {
                    //
                    // NOTE: Get the identity of the user for the current thread.
                    //
                    WindowsIdentity identity = WindowsIdentity.GetCurrent();

                    //
                    // NOTE: Attempt to get the file security object for this file or
                    //       directory.
                    //
                    FileSystemSecurity security;

                    //
                    // NOTE: Use the correct method based on whether this path represents
                    //       a file or directory.
                    //
                    if (Directory.Exists(path))
                        security = Directory.GetAccessControl(path);
                    else
                        security = File.GetAccessControl(path);

                    //
                    // NOTE: Attempt to get the access rules for this file or directory.
                    //
                    AuthorizationRuleCollection accessRules =
                        security.GetAccessRules(true, true, typeof(SecurityIdentifier));

                    //
                    // NOTE: Iterate over the access rules and change the granted rights
                    //       according to what we find.
                    //
                    foreach (FileSystemAccessRule accessRule in accessRules)
                    {
                        //
                        // BUGBUG: Double-check this logic.  There appears to be no formal
                        //         documentation in MSDN that indicates the correct way to
                        //         perform an access check using the .NET Framework supplied
                        //         functionality; however, this works and appears to be
                        //         "correct".
                        //
                        // NOTE: Check to see if this access rule applies to the current
                        //       user directly or indirectly via a group that the current
                        //       user belongs to.
                        //
                        if (identity.User.Equals(accessRule.IdentityReference) ||
                            identity.Groups.Contains(accessRule.IdentityReference))
                        {
                            switch (accessRule.AccessControlType)
                            {
                                case AccessControlType.Allow:
                                    //
                                    // NOTE: The file system rights access mask represents
                                    //       rights the user has been granted, add them to
                                    //       the granted rights.
                                    //
                                    grantedRights |= accessRule.FileSystemRights;
                                    break;
                                case AccessControlType.Deny:
                                    //
                                    // NOTE: The file system rights access mask represents
                                    //       rights the user has been deined, remove them
                                    //       from the granted rights.
                                    //
                                    grantedRights &= ~accessRule.FileSystemRights;
                                    break;
                            }
                        }
                    }

                    //
                    // NOTE: Only return "true" for the access status if ALL the desired
                    //       rights are granted.
                    //
                    accessStatus = HasRights(grantedRights, desiredRights, true);
                }

                return ReturnCode.Ok;
            }
            catch (Exception e)
            {
                error = e;
                return ReturnCode.Error;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static long GetFileSize(
            string path
            )
        {
            try
            {
                FileInfo fileInfo = new FileInfo(path); /* throw */

                return fileInfo.Length; /* throw */
            }
            catch
            {
                // do nothing.
            }

            return 0;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static ReturnCode CheckReadOnlyAndDirectory(
            string path,
            ref bool exists,
            ref bool readOnly,
            ref bool directory,
            ref Result error
            )
        {
            try
            {
                FileAttributes fileAttributes = (FileAttributes)0; /* NONE */

                bool newExists = PathOps.PathExists(path);

                if (newExists)
                    fileAttributes = File.GetAttributes(path);

                exists = newExists; /* NOTE: Transactional. */
                readOnly = ((fileAttributes & FileAttributes.ReadOnly) == FileAttributes.ReadOnly);
                directory = ((fileAttributes & FileAttributes.Directory) == FileAttributes.Directory);

                return ReturnCode.Ok;
            }
            catch (Exception e)
            {
                error = String.Format(
                    "can't get attributes \"{0}\": {1}",
                    path, e.Message);

                return ReturnCode.Error;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode VerifyPath(
            string path,
            FilePermission permissions,
            ref Result error
            )
        {
            if (!String.IsNullOrEmpty(path))
            {
                bool exists = false;
                bool readOnly = false;
                bool directory = false;

                if (CheckReadOnlyAndDirectory(path, ref exists, ref readOnly, ref directory, ref error) == ReturnCode.Ok)
                {
                    if ((directory && FlagOps.HasFlags(permissions, FilePermission.Directory, true)) ||
                        (!directory && FlagOps.HasFlags(permissions, FilePermission.File, true)) ||
                        !FlagOps.HasFlags(permissions, FilePermission.Directory | FilePermission.File, false))
                    {
                        //
                        // NOTE: For directories, use 1 as the size so that we do not allow it to be
                        //       verified if it exists and the caller indicated that it must not exist.
                        //
                        //       For our purposes, we want to consider zero byte files to "not exist"
                        //       because of how temporary file names are allocated.
                        //
                        long fileSize = !directory ? GetFileSize(path) : 1;

                        if ((exists && FlagOps.HasFlags(permissions, FilePermission.Exists, true)) ||
                            ((!exists || (fileSize == 0)) && FlagOps.HasFlags(permissions, FilePermission.NotExists, true)) ||
                            !FlagOps.HasFlags(permissions, FilePermission.Exists | FilePermission.NotExists, false))
                        {
                            //
                            // NOTE: Are we asking for write permission on the file?
                            //
                            bool write = FlagOps.HasFlags(permissions, FilePermission.Write, true);

                            if (!write || !readOnly)
                            {
                                FileSystemRights desiredRights =
                                    FilePermissionsToFileSystemRights(permissions);

                                if (desiredRights != (FileSystemRights)0)
                                {
                                    if (exists || write)
                                    {
                                        string originalPath = path; /* save */
                                        FileSystemRights grantedRights = (FileSystemRights)0; /* None */
                                        bool accessStatus = false;

                                        //
                                        // NOTE: If the file does not exist, be sure we can create
                                        //       things in the parent directory.
                                        //
                                        if (!exists)
                                            path = Path.GetDirectoryName(path);

                                        if (!String.IsNullOrEmpty(path))
                                        {
                                            if (AccessCheck(path, desiredRights, ref grantedRights, ref accessStatus, ref error) == ReturnCode.Ok)
                                            {
                                                if (accessStatus)
                                                    return ReturnCode.Ok;
                                                else
                                                    error = String.Format(
                                                        "access denied: no rights to path \"{0}\"",
                                                        path);
                                            }
                                        }
                                        else
                                        {
                                            error = String.Format(
                                                "no directory in path \"{0}\"",
                                                originalPath);
                                        }
                                    }
                                    else
                                    {
                                        error = String.Format(
                                            "access denied: no such path \"{0}\"",
                                            path);
                                    }
                                }
                                else
                                {
                                    return ReturnCode.Ok;
                                }
                            }
                            else
                            {
                                error = String.Format(
                                    "access denied: path \"{0}\" is read-only.",
                                    path);
                            }
                        }
                        else
                        {
                            error = String.Format(exists ?
                                "path already exists \"{0}\"" : "no such path \"{0}\"",
                                path);
                        }
                    }
                    else
                    {
                        error = String.Format(
                            "no such path \"{0}\"",
                            path);
                    }
                }
            }
            else
            {
                error = "invalid path";
            }

            return ReturnCode.Error;
        }
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static string GetFileType(
            string path
            )
        {
            if (Directory.Exists(path))
                return "directory";
            else if (File.Exists(path))
                return "file";
            else
                return "unknown";
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static FileAccess FileAccessFromAccess(
            MapOpenAccess access
            )
        {
            if (FlagOps.HasFlags(access, MapOpenAccess.RdWr, true))
                return FileAccess.ReadWrite;

            if (FlagOps.HasFlags(access, MapOpenAccess.WrOnly, true))
                return FileAccess.Write;

            return FileAccess.Read;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static FileMode FileModeFromAccess(
            MapOpenAccess access
            )
        {
            //
            // NOTE: The MapOpenAccess.Append flag is not handled by this function
            //       as it requires special treatment to get POSIX-like behavior.
            //
            switch (access & (MapOpenAccess.Creat | MapOpenAccess.Excl | MapOpenAccess.Trunc))
            {
                case (MapOpenAccess.Creat | MapOpenAccess.Excl):
                case (MapOpenAccess.Creat | MapOpenAccess.Excl | MapOpenAccess.Trunc):
                    //
                    // NOTE: Create new file, error if file exists.
                    //
                    return FileMode.CreateNew;

                case (MapOpenAccess.Creat | MapOpenAccess.Trunc):
                    //
                    // NOTE: Create new file, overwrite if file exists.
                    //
                    return FileMode.Create;

                case MapOpenAccess.Creat:
                    //
                    // NOTE: Create a new file or open existing file.
                    //
                    return FileMode.OpenOrCreate;

                case MapOpenAccess.Trunc:
                case (MapOpenAccess.Trunc | MapOpenAccess.Excl):
                    //
                    // NOTE: Open existing file and truncate to zero length,
                    //       error if file does not exist.
                    //
                    return FileMode.Truncate;
            }

            //
            // NOTE: Open existing file, error if file does not exist.
            //
            return FileMode.Open;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static bool HaveFileAttributes(
            FileAttributes attributes,
            FileAttributes haveAttributes
            )
        {
            return ((attributes & haveAttributes) == haveAttributes);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode SetFileAttributes(
            string path,
            FileAttributes fileAttributes,
            ref Result error
            )
        {
            try
            {
                File.SetAttributes(path, fileAttributes);

                return ReturnCode.Ok;
            }
            catch (Exception e)
            {
                error = String.Format(
                    "can't set attributes \"{0}\": {1}",
                    path, e.Message);

                return ReturnCode.Error;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode GetFileAttributes(
            string path,
            ref FileAttributes fileAttributes,
            ref Result error
            )
        {
            try
            {
                fileAttributes = File.GetAttributes(path);

                return ReturnCode.Ok;
            }
            catch (Exception e)
            {
                error = String.Format(
                    "can't get attributes \"{0}\": {1}",
                    path, e.Message);

                return ReturnCode.Error;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode FileDelete(
            IList paths,
            bool recursive,
            bool force,
            ref Result error
            )
        {
            if ((paths != null) && (paths.Count > 0))
            {
                for (int index = 0; index < paths.Count; index++)
                {
                    if (paths[index] != null)
                    {
                        string path = paths[index].ToString();

                        try
                        {
                            if (File.Exists(path))
                            {
                                if (force)
                                {
                                    FileAttributes fileAttributes = File.GetAttributes(path);

                                    if ((fileAttributes & FileAttributes.ReadOnly) == FileAttributes.ReadOnly)
                                    {
                                        fileAttributes &= ~FileAttributes.ReadOnly;
                                        File.SetAttributes(path, fileAttributes);
                                    }
                                }

                                File.Delete(path);
                            }
                            else if (Directory.Exists(path))
                            {
                                if (recursive && force)
                                {
                                    string[] fileNames = Directory.GetFiles(
                                        path, Characters.Asterisk.ToString(),
                                        SearchOption.AllDirectories);

                                    foreach (string fileName in fileNames)
                                    {
                                        FileAttributes fileAttributes = File.GetAttributes(fileName);

                                        if ((fileAttributes & FileAttributes.ReadOnly) == FileAttributes.ReadOnly)
                                        {
                                            fileAttributes &= ~FileAttributes.ReadOnly;
                                            File.SetAttributes(fileName, fileAttributes);
                                        }
                                    }
                                }

                                Directory.Delete(path, recursive);
                            }
                            else
                            {
                                error = String.Format(
                                    "error deleting \"{0}\": no such file or directory",
                                    path);

                                return ReturnCode.Error;
                            }
                        }
                        catch (Exception e)
                        {
                            error = String.Format(
                                "error deleting \"{0}\": {1}",
                                path, e.Message);

                            return ReturnCode.Error;
                        }
                    }
                    else
                    {
                        error = "invalid file name";

                        return ReturnCode.Error;
                    }
                }

                return ReturnCode.Ok;
            }
            else
            {
                error = "no files";

                return ReturnCode.Error;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode FileCopy(
            IList fileNames,
            string path,
            bool move,
            bool force,
            ref Result error
            )
        {
            if ((fileNames != null) && (fileNames.Count > 0))
            {
                if (!String.IsNullOrEmpty(path))
                {
                    if (fileNames.Count > 1)
                    {
                        if (Directory.Exists(path))
                        {
                            for (int index = 0; index < fileNames.Count; index++)
                            {
                                if (fileNames[index] != null)
                                {
                                    string fileName = fileNames[index].ToString();

                                    if (File.Exists(fileName))
                                    {
                                        try
                                        {
                                            File.Copy(fileName, PathOps.CombinePath(
                                                false, path, Path.GetFileName(fileName)), force);

                                            if (move)
                                                File.Delete(fileName);
                                        }
                                        catch (Exception e)
                                        {
                                            error = String.Format(
                                                "error {0} \"{1}\": {2}",
                                                move ? "moving" : "copying", fileName, e.Message);

                                            return ReturnCode.Error;
                                        }
                                    }
                                    else
                                    {
                                        error = String.Format(
                                            "error {0} \"{1}\": no such file",
                                            move ? "moving" : "copying", fileName);

                                        return ReturnCode.Error;
                                    }
                                }
                                else
                                {
                                    error = "invalid file name";

                                    return ReturnCode.Error;
                                }
                            }

                            return ReturnCode.Ok;
                        }
                        else
                        {
                            error = String.Format(
                                "error {0}: target \"{1}\" is not a directory",
                                move ? "moving" : "copying", path);

                            return ReturnCode.Error;
                        }
                    }
                    else
                    {
                        if (fileNames[0] != null)
                        {
                            string fileName = fileNames[0].ToString();

                            try
                            {
                                File.Copy(fileName, path, force);

                                if (move)
                                    File.Delete(fileName);

                                return ReturnCode.Ok;
                            }
                            catch (Exception e)
                            {
                                error = String.Format(
                                    "error {0} \"{1}\": {2}",
                                    move ? "moving" : "copying", fileName, e.Message);

                                return ReturnCode.Error;
                            }
                        }
                        else
                        {
                            error = "invalid file name";

                            return ReturnCode.Error;
                        }
                    }
                }
                else
                {
                    error = "invalid target";

                    return ReturnCode.Error;
                }
            }
            else
            {
                error = "no files";

                return ReturnCode.Error;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode GetFileTime(
            string path,
            GetFileDateTimeCallback callback,
            ref DateTime dateTime,
            ref Result error
            )
        {
            try
            {
                dateTime = callback(path);

                return ReturnCode.Ok;
            }
            catch (Exception e)
            {
                error = String.Format(
                    "can't get file time \"{0}\": {1}",
                    path, e.Message);

                return ReturnCode.Error;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode SetFileTime(
            string path,
            SetFileDateTimeCallback callback,
            DateTime dateTime,
            ref Result error
            )
        {
            try
            {
                callback(path, dateTime);

                return ReturnCode.Ok;
            }
            catch (Exception e)
            {
                error = String.Format(
                    "can't set file time \"{0}\": {1}",
                    path, e.Message);

                return ReturnCode.Error;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode GetFileSystemEntries(
            string path,
            string searchPattern,
            ref StringList entries,
            ref Result error
            )
        {
            if (Directory.Exists(path))
            {
                try
                {
                    if (searchPattern != null)
                        entries = new StringList(
                            PathOps.GetUnixPath,
                            Directory.GetFileSystemEntries(path, searchPattern));
                    else
                        entries = new StringList(
                            PathOps.GetUnixPath,
                            Directory.GetFileSystemEntries(path));

                    return ReturnCode.Ok;
                }
                catch (Exception e)
                {
                    error = String.Format(
                        "can't get entries \"{0}\": {1}",
                        path, e.Message);

                    return ReturnCode.Error;
                }
            }
            else
            {
                error = String.Format(
                    "can't get entries \"{0}\": no such directory",
                    path);

                return ReturnCode.Error;
            }
        }
    }
}
