/*
 * SocketOps.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

using System;
using System.ComponentModel;
using System.Globalization;
using System.Net;
using System.Net.NetworkInformation;
using System.Net.Security;
using System.Net.Sockets;
using System.Reflection;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using Eagle._Attributes;
using Eagle._Components.Public;
using Eagle._Constants;
using Eagle._Containers.Public;
using Eagle._Interfaces.Public;

namespace Eagle._Components.Private
{
    [ObjectId("71b14766-48a0-45d5-9254-640fde03509d")]
    internal static class SocketOps
    {
        #region Private Static Data
        private static readonly object syncRoot = new object();

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static bool useLegacyCertificatePolicy = false;

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static ICertificatePolicy savedCertificatePolicy;
        private static bool haveSavedCertificatePolicy;

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static readonly ICertificatePolicy softwareUpdateCertificatePolicy =
            new SoftwareUpdateCertificatePolicy();

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static PropertyInfo tcpListenerActive;
        private static PropertyInfo socketCleanedUp;
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Software Update Certificate Policy Class
        [ObjectId("4062e197-ed96-4db3-87e8-f463e5fb818b")]
        private sealed class SoftwareUpdateCertificatePolicy : ICertificatePolicy
        {
            #region ICertificatePolicy Members
            public bool CheckValidationResult(
                ServicePoint srvPoint,
                X509Certificate certificate,
                WebRequest request,
                int certificateProblem
                )
            {
                return (certificateProblem == 0) ||
                    IsSoftwareUpdateCertificate(certificate);
            }
            #endregion
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Static Download Helper Methods
        private static StringList GetAsyncCompletedArguments(
            Uri uri,
            string fileName,
            AsyncCompletedEventArgs eventArgs
            )
        {
            StringList result = new StringList();

            if (uri != null)
            {
                result.Add("uri");
                result.Add(uri.ToString());
            }

            if (fileName != null)
            {
                result.Add("fileName");
                result.Add(fileName);
            }

            if (eventArgs != null)
            {
                bool canceled = eventArgs.Cancelled;

                result.Add("canceled");
                result.Add(canceled.ToString());

                Exception exception = eventArgs.Error;

                if (exception != null)
                {
                    result.Add("exception");
                    result.Add(exception.GetType().ToString());
                    result.Add("error");
                    result.Add(exception.ToString());
                }
            }

            return result;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Static Download Event Handlers
        private static void DownloadAsyncCompleted(
            object sender,
            AsyncCompletedEventArgs e
            )
        {
            try
            {
                if (e == null)
                    return;

                ICallback callback = e.UserState as ICallback;

                if (callback == null)
                    return;

                Uri uri = null;
                string fileName = null;
                IClientData clientData = callback.ClientData;

                if (clientData != null)
                {
                    IAnyTriplet<WebClient, Uri, string> triplet =
                        clientData.Data as IAnyTriplet<WebClient, Uri, string>;

                    if (triplet != null)
                    {
                        WebClient webClient = triplet.X;

                        if (webClient != null)
                        {
                            webClient.Dispose();
                            webClient = null;
                        }

                        uri = triplet.Y;
                        fileName = triplet.Z;
                    }

                    clientData.Data = null;
                }

                ReturnCode code;
                Result result = null;

                code = callback.Invoke(
                    GetAsyncCompletedArguments(uri, fileName, e), ref result);

                if (code != ReturnCode.Ok)
                    DebugOps.Complain(code, result);
            }
            catch (Exception ex)
            {
                DebugOps.Complain(ReturnCode.Error, ex);
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Public Web Download Methods
        public static ReturnCode DownloadData(
            Uri uri,
            bool trusted,
            ref byte[] bytes,
            ref Result error
            )
        {
            bool wasTrusted = IsSoftwareUpdateTrusted();

            RuntimeOps.DebugTrace(String.Format(
                "{0}: DownloadData: uri = {1}, trusted = {2}, wasTrusted = {3}{4}",
                GlobalState.GetCurrentThreadId(), FormatOps.WrapOrNull(uri),
                trusted, wasTrusted, Environment.NewLine), typeof(SocketOps).Name);

            ReturnCode code = trusted ? wasTrusted ? ReturnCode.Ok :
                SetSoftwareUpdateTrusted(trusted, ref error) : ReturnCode.Ok;

            if (code == ReturnCode.Ok)
            {
                try
                {
                    try
                    {
                        using (WebClient webClient = new WebClient())
                        {
                            bytes = webClient.DownloadData(uri);
                        }

                        return ReturnCode.Ok;
                    }
                    catch (Exception e)
                    {
                        error = e;
                    }
                }
                finally
                {
                    if (trusted && !wasTrusted)
                    {
                        ReturnCode untrustCode;
                        Result untrustError = null;

                        untrustCode = SetSoftwareUpdateTrusted(
                            false, ref untrustError);

                        if (untrustCode != ReturnCode.Ok)
                            DebugOps.Complain(untrustCode, untrustError);
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode DownloadFile(
            Uri uri,
            string fileName,
            bool trusted,
            ref Result error
            )
        {
            bool wasTrusted = IsSoftwareUpdateTrusted();

            RuntimeOps.DebugTrace(String.Format(
                "{0}: DownloadFile: uri = {1}, fileName = {2}, trusted = {3}, wasTrusted = {4}{5}",
                GlobalState.GetCurrentThreadId(), FormatOps.WrapOrNull(uri),
                FormatOps.WrapOrNull(fileName), trusted, wasTrusted, Environment.NewLine),
                typeof(SocketOps).Name);

            ReturnCode code = trusted ? wasTrusted ? ReturnCode.Ok :
                SetSoftwareUpdateTrusted(trusted, ref error) : ReturnCode.Ok;

            if (code == ReturnCode.Ok)
            {
                try
                {
                    try
                    {
                        using (WebClient webClient = new WebClient())
                        {
                            webClient.DownloadFile(uri, fileName);
                        }

                        return ReturnCode.Ok;
                    }
                    catch (Exception e)
                    {
                        error = e;
                    }
                }
                finally
                {
                    if (trusted && !wasTrusted)
                    {
                        ReturnCode untrustCode;
                        Result untrustError = null;

                        untrustCode = SetSoftwareUpdateTrusted(
                            false, ref untrustError);

                        if (untrustCode != ReturnCode.Ok)
                            DebugOps.Complain(untrustCode, untrustError);
                    }
                }
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode DownloadFileAsync(
            Interpreter interpreter,
            StringList arguments,
            CallbackFlags flags,
            Uri uri,
            string fileName,
            ref Result error
            )
        {
            RuntimeOps.DebugTrace(String.Format(
                "{0}: DownloadFileAsync: interpreter = {1}, " +
                "arguments = {2}, flags = {3}, uri = {4}, " +
                "fileName = {5}{6}",
                GlobalState.GetCurrentThreadId(),
                FormatOps.WrapOrNull(interpreter),
                FormatOps.WrapOrNull(true, true, arguments),
                FormatOps.WrapOrNull(flags),
                FormatOps.WrapOrNull(uri),
                FormatOps.WrapOrNull(fileName),
                Environment.NewLine), typeof(SocketOps).Name);

            ReturnCode code = ReturnCode.Ok;
            WebClient webClient = null;

            try
            {
                ICallback callback = CommandCallback.Create(
                    flags, interpreter, null, arguments, ref error);

                if (callback != null)
                {
                    try
                    {
                        webClient = new WebClient();

                        callback.ClientData = new ClientData(
                            new AnyTriplet<WebClient, Uri, string>(
                                webClient, uri, fileName));

                        webClient.DownloadFileCompleted +=
                            new AsyncCompletedEventHandler(
                                DownloadAsyncCompleted);

                        webClient.DownloadFileAsync(uri, fileName, callback);
                    }
                    catch (Exception e)
                    {
                        error = e;
                        code = ReturnCode.Error;
                    }
                }
                else
                {
                    code = ReturnCode.Error;
                }
            }
            finally
            {
                if ((code != ReturnCode.Ok) && (webClient != null))
                {
                    ReturnCode disposeCode;
                    Result disposeError = null;

                    disposeCode = RuntimeOps.TryDisposeObject(
                        webClient, ref disposeError);

                    if (disposeCode != ReturnCode.Ok)
                        DebugOps.Complain(disposeCode, disposeError);
                }
            }

            return code;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Public Software Update Trust Methods
        public static bool IsSoftwareUpdateTrusted()
        {
            return !useLegacyCertificatePolicy && !CommonOps.Runtime.IsMono() ?
                (ServicePointManager.ServerCertificateValidationCallback ==
                    SoftwareUpdateRemoteCertificateValidationCallback) :
                Object.ReferenceEquals(ServicePointManager.CertificatePolicy,
                    softwareUpdateCertificatePolicy);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode SetSoftwareUpdateTrusted(
            bool trusted,
            ref Result error
            )
        {
            try
            {
                if (!useLegacyCertificatePolicy &&
                    !CommonOps.Runtime.IsMono())
                {
                    //
                    // NOTE: When using the .NET Framework, use the newer
                    //       certification validation callback interface.
                    //
                    if (trusted)
                        ServicePointManager.ServerCertificateValidationCallback +=
                            SoftwareUpdateRemoteCertificateValidationCallback;
                    else
                        ServicePointManager.ServerCertificateValidationCallback -=
                            SoftwareUpdateRemoteCertificateValidationCallback;

                    //
                    // NOTE: Report this trust change to any trace listeners.
                    //
                    RuntimeOps.DebugTrace(String.Format(
                        "{0}: SetSoftwareUpdateTrusted: {1} " +
                        "SoftwareUpdateRemoteCertificateValidationCallback{2}",
                        GlobalState.GetCurrentThreadId(),
                        trusted ? "added" : "removed", Environment.NewLine),
                        typeof(SocketOps).Name);
                }
                else
                {
                    //
                    // NOTE: When using Mono, fallback to the "obsolete"
                    //       CertificatePolicy property.
                    //
                    if (trusted)
                    {
                        //
                        // NOTE: Save the current ICertificatePolicy for
                        //       possible later restoration.
                        //
                        savedCertificatePolicy =
                            ServicePointManager.CertificatePolicy;

                        haveSavedCertificatePolicy = true;

                        //
                        // NOTE: Set the ICertificatePolicy to the one we
                        //       use for software updates.
                        //
                        ServicePointManager.CertificatePolicy =
                            softwareUpdateCertificatePolicy;
                    }
                    else
                    {
                        //
                        // NOTE: Restore the previously saved
                        //       ICertificatePolicy, if any.
                        //
                        if (haveSavedCertificatePolicy)
                        {
                            //
                            // NOTE: Restore the saved ICertificatePolicy.
                            //
                            ServicePointManager.CertificatePolicy =
                                savedCertificatePolicy;

                            //
                            // NOTE: Clear the saved ICertificatePolicy.
                            //
                            savedCertificatePolicy = null;
                            haveSavedCertificatePolicy = false;
                        }
                    }

                    //
                    // NOTE: Report this trust change to any trace listeners.
                    //
                    RuntimeOps.DebugTrace(String.Format(
                        "{0}: SetSoftwareUpdateTrusted: {1} " +
                        "SoftwareUpdateCertificatePolicy{2}",
                        GlobalState.GetCurrentThreadId(),
                        trusted ? "overridden" : "restored", Environment.NewLine),
                        typeof(SocketOps).Name);
                }

                return ReturnCode.Ok;
            }
            catch (Exception e)
            {
                error = e;
            }

            //
            // NOTE: Report this trust change to any trace listeners.
            //
            RuntimeOps.DebugTrace(String.Format(
                "{0}: SetSoftwareUpdateTrusted: trusted = {1}, error = {2}{3}",
                GlobalState.GetCurrentThreadId(), trusted,
                FormatOps.WrapOrNull(error), Environment.NewLine),
                typeof(SocketOps).Name);

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Software Update Trust Methods
        private static bool SoftwareUpdateRemoteCertificateValidationCallback(
            object sender,
            X509Certificate certificate,
            X509Chain chain,
            SslPolicyErrors sslPolicyErrors
            )
        {
            return (sslPolicyErrors == SslPolicyErrors.None) ||
                IsSoftwareUpdateCertificate(certificate);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static bool IsSoftwareUpdateCertificate(
            X509Certificate certificate
            )
        {
            bool result;

            //
            // NOTE: Make sure the certificate public key matches what
            //       we expect it to be for our own software updates.
            //
            if (certificate != null)
            {
                //
                // NOTE: Grab the public key of the certificate.
                //
                byte[] certificatePublicKey = certificate.GetPublicKey();

                //
                // NOTE: Compare the public key of the certificate to
                //       the one we trust for our software updates.
                //
                result = RuntimeOps.ArrayEquals(
                    certificatePublicKey, _Constants.PublicKey.SoftwareUpdate);
            }
            else
            {
                result = false;
            }

            //
            // NOTE: Report this trust result to any trace listeners.
            //
            RuntimeOps.DebugTrace(String.Format(
                "{0}: IsSoftwareUpdateCertificate: certificate = {1}, " +
                "result = {2}{3}", GlobalState.GetCurrentThreadId(),
                FormatOps.Certificate(certificate, false, true), result,
                Environment.NewLine), typeof(SocketOps).Name);

            return result;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Public Diagnostic Methods
        public static ReturnCode Ping(
            string host,
            int timeout,
            ref IPStatus status,
            ref long roundtripTime,
            ref Result error
            )
        {
            try
            {
                using (Ping ping = new Ping())
                {
                    PingReply reply = ping.Send(host, timeout);

                    //
                    // NOTE: Populate reply information for the caller.
                    //
                    status = reply.Status;
                    roundtripTime = reply.RoundtripTime;
                }

                return ReturnCode.Ok;
            }
            catch (Exception e)
            {
                //
                // NOTE: Populate error information for the caller.
                //
                error = e;
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Public Network Helper Methods
        public static IPAddress GetIpAddress(
            string host,
            bool strict,
            ref Result error
            )
        {
            IPAddress result = null;

            if (!String.IsNullOrEmpty(host))
            {
                if (!IPAddress.TryParse(host, out result))
                {
                    try
                    {
                        //
                        // NOTE: Attempt to resolve the host name to one or
                        //       more IP addresses. This is required even for
                        //       things like "localhost", etc.
                        //
                        IPAddress[] addresses = Dns.GetHostAddresses(host);

                        if (addresses != null)
                        {
                            //
                            // NOTE: We require an IPv4 address.
                            //
                            for (int index = 0; index < addresses.Length; index++)
                            {
                                if (addresses[index].AddressFamily == AddressFamily.InterNetwork)
                                {
                                    result = addresses[index];
                                    break;
                                }
                            }
                        }
                    }
                    catch (Exception e)
                    {
                        error = e;
                    }
                }
            }
            else
            {
                if (strict)
                    error = "invalid host name or address";
                else
                    result = IPAddress.Any;
            }

            return result;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static int GetPortNumber(
            string port,
            CultureInfo cultureInfo,
            bool strict,
            ref Result error
            )
        {
            int result = Port.Invalid;

            if (!String.IsNullOrEmpty(port))
            {
                if (Value.GetInteger2(port, ValueFlags.AnyInteger, cultureInfo,
                        ref result, ref error) == ReturnCode.Ok)
                {
                    return result;
                }

                //
                // FIXME: Else we should try to lookup the service name using getservbyname()
                //        API; however, the .NET Framework does not expose this functionality.
                //        A possible workaround would be to see if the necessary DLL is already
                //        loaded (e.g. WS2_32.DLL) and simply P/Invoke the function ourselves.
                //
            }
            else if (!strict)
            {
                result = Port.Automatic;
            }

            return result;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Public Network Client Methods
        public static TcpClient NewTcpClient(
            string localAddress,
            string localPort,
            CultureInfo cultureInfo,
            ref Result error
            )
        {
            IPAddress address = GetIpAddress(localAddress, false, ref error);

            if (address != null)
            {
                int port = GetPortNumber(localPort, cultureInfo, false, ref error);

                if (port != Port.Invalid)
                    return new TcpClient(new IPEndPoint(address, (int)port));
            }

            return null;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static ReturnCode Connect(
            TcpClient client,
            string remoteHost,
            string remotePort,
            CultureInfo cultureInfo,
            ref Result error
            )
        {
            IPAddress address = GetIpAddress(remoteHost, true, ref error);

            if (address != null)
            {
                int port = GetPortNumber(remotePort, cultureInfo, true, ref error);

                if (port != Port.Invalid)
                {
                    try
                    {
                        client.Connect(new IPEndPoint(address, (int)port));

                        return ReturnCode.Ok;
                    }
                    catch (Exception e)
                    {
                        error = e;
                    }
                }
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Public Network Server Methods
        public static ReturnCode GetServerScript(
            TcpClient client,
            string channelId,
            string oldScript,
            ref string newScript,
            ref Result error
            )
        {
            try
            {
                if (client != null)
                {
                    Socket socket = client.Client;

                    if (socket != null)
                    {
                        IPEndPoint endPoint = socket.RemoteEndPoint as IPEndPoint;

                        if (endPoint != null)
                        {
                            StringBuilder builder = StringOps.NewStringBuilder(oldScript);

                            builder.Append(Characters.Space);
                            builder.Append(channelId);
                            builder.Append(Characters.Space);
                            builder.Append(endPoint.Address);
                            builder.Append(Characters.Space);
                            builder.Append(endPoint.Port);

                            newScript = builder.ToString();

                            return ReturnCode.Ok;
                        }
                        else
                        {
                            error = "invalid remote endpoint";
                        }
                    }
                    else
                    {
                        error = "invalid client socket";
                    }
                }
                else
                {
                    error = "invalid client";
                }
            }
            catch (Exception e)
            {
                error = e;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static bool IsListenerActive(
            TcpListener listener
            )
        {
            try
            {
                #region Static Lock Held
                lock (syncRoot)
                {
                    //
                    // HACK: Why must we do this?  This member is marked as "protected";
                    //       however, we need to know this information and we do not want
                    //       to derive a custom class to get it; therefore, just use
                    //       Reflection.  We cache the PropertyInfo object so that we do
                    //       not need to look it up more than once.
                    //
                    if (tcpListenerActive == null)
                        tcpListenerActive = typeof(TcpListener).GetProperty(
                            "Active", BindingFlags.Instance | BindingFlags.NonPublic);

                    if ((tcpListenerActive != null) && (listener != null))
                        return (bool)tcpListenerActive.GetValue(listener, null);
                }
                #endregion
            }
            catch
            {
                // do nothing.
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static bool IsCleanedUp(
            Socket socket
            )
        {
            try
            {
                #region Static Lock Held
                lock (syncRoot)
                {
                    //
                    // HACK: Why must we do this?  This member is marked as "protected";
                    //       however, we need to know this information and we do not want
                    //       to derive a custom class to get it; therefore, just use
                    //       Reflection.  We cache the PropertyInfo object so that we do
                    //       not need to look it up more than once.
                    //
                    if (socketCleanedUp == null)
                        socketCleanedUp = typeof(Socket).GetProperty(
                            "CleanedUp", BindingFlags.Instance | BindingFlags.NonPublic);

                    if ((socketCleanedUp != null) && (socket != null))
                        return (bool)socketCleanedUp.GetValue(socket, null);
                }
                #endregion
            }
            catch
            {
                // do nothing.
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public static TcpListener NewTcpListener(
            string localAddress,
            string localPort,
            CultureInfo cultureInfo,
            ref Result error
            )
        {
            IPAddress address = null;

            if (localAddress != null)
                address = GetIpAddress(localAddress, true, ref error);

            if ((localAddress == null) || (address != null))
            {
                int port = GetPortNumber(localPort, cultureInfo, true, ref error);

                if (port != Port.Invalid)
                    return (address != null) ?
                        new TcpListener(address, port) : new TcpListener(port);
            }

            return null;
        }
        #endregion
    }
}

