/*
 * Console.cs --
 *
 * Copyright (c) 2007-2012 by Joe Mistachkin.  All rights reserved.
 *
 * See the file "license.terms" for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * RCS: @(#) $Id: $
 */

#if !CONSOLE
#error "This file cannot be compiled or used properly with console support disabled."
#endif

using System;
using System.IO;

#if NATIVE && WINDOWS
using System.Runtime.InteropServices;
#endif

using System.Text;
using System.Threading;
using Eagle._Attributes;
using Eagle._Components.Public;
using Eagle._Components.Private;
using Eagle._Constants;
using Eagle._Containers.Public;
using Eagle._Interfaces.Public;

#if DRAWING
using System.Drawing;
#endif

namespace Eagle._Hosts
{
    [ObjectId("e15283cf-00b4-44f2-a16e-48cf061e53d1")]
    public class Console : Core, IDisposable
    {
        #region Private Data
        private static int referenceCount = 0;

        ///////////////////////////////////////////////////////////////////////////////////////////////

#if NATIVE && WINDOWS && DRAWING
        private Icon icon;
        private IntPtr oldBigIcon;
        private IntPtr oldSmallIcon;
#endif

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private string savedTitle;

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private ConsoleColor savedForegroundColor = _ConsoleColor.None;
        private ConsoleColor savedBackgroundColor = _ConsoleColor.None;

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private int savedWindowWidth = _Size.Invalid;
        private int savedWindowHeight = _Size.Invalid;
        private int savedBufferWidth = _Size.Invalid;
        private int savedBufferHeight = _Size.Invalid;
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Private Constants
        //
        // HACK: These are considered to be "best guess" values.
        //       Please adjust them to suit your taste as necessary.
        //
        private static readonly int MinimumWindowWidth = 40;
        private static readonly int CompactWindowWidth = 80;
        private static readonly int FullWindowWidth = 120;
        private static readonly int SuperFullWindowWidth = 160;
        private static readonly int JumboWindowWidth = 200;
        private static readonly int SuperJumboWindowWidth = 230;

        private static readonly int MinimumWindowHeight = 10;
        private static readonly int CompactWindowHeight = 25;
        private static readonly int FullWindowHeight = 40;
        private static readonly int SuperFullWindowHeight = 60;
        private static readonly int JumboWindowHeight = 75;
        private static readonly int SuperJumboWindowHeight = 90;

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private const string TitleFormat = "{0}{1} {2}{3}";
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Public Constructors
        public Console(
            IHostData hostData
            )
            : base(hostData)
        {
            //
            // NOTE: Save the original buffer and window sizes.
            //
            /* IGNORED */
            SaveSize();

            //
            // NOTE: Save the original colors.
            //
            /* IGNORED */
            SaveColors();

            /* IGNORED */
            Setup(this, true, false);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Console Handling
        #region System.Console Size Handling
        protected override int WindowWidth
        {
            get
            {
                try
                {
                    return System.Console.WindowWidth;
                }
                catch
                {
                    return base.WindowWidth;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected override int WindowHeight
        {
            get
            {
                try
                {
                    return System.Console.WindowHeight;
                }
                catch
                {
                    return base.WindowHeight;
                }
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual bool SaveSize()
        {
            //
            // NOTE: Save original console dimensions in case we need to restore
            //       from the later (e.g. ResetSize).
            //
            try
            {
                savedBufferWidth = System.Console.BufferWidth;
                savedBufferHeight = System.Console.BufferHeight;

                savedWindowWidth = System.Console.WindowWidth;
                savedWindowHeight = System.Console.WindowHeight;

                return true;
            }
            catch
            {
                return false;
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region System.Console Color Save/Restore
        protected virtual bool SaveColors()
        {
            //
            // NOTE: Save original console colors in case we need to restore
            //       from the later.
            //
            try
            {
                return GetColors(ref savedForegroundColor, ref savedBackgroundColor);
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual bool RestoreColors()
        {
            //
            // NOTE: Restore the originally saved console colors.
            //
            try
            {
                return SetColors(true, true, savedForegroundColor, savedBackgroundColor);
            }
            catch
            {
                return false;
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region System.Console Title Handling
        protected virtual bool SaveTitle(
            ref Result error
            )
        {
            try
            {
                if (savedTitle == null)
                    savedTitle = System.Console.Title;

                return true;
            }
            catch (Exception e)
            {
                error = e;
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual bool RestoreTitle(
            ref Result error
            )
        {
            try
            {
                if (savedTitle != null)
                {
                    System.Console.Title = savedTitle;
                    savedTitle = null;
                }

                return true;
            }
            catch (Exception e)
            {
                error = e;
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected virtual bool SetupTitle(
            bool setup
            )
        {
            try
            {
                //
                // NOTE: Has changing the title been explicitly disabled?
                //
                if (!NoTitle && IsOpen())
                {
                    Interpreter interpreter = SafeGetInterpreter();
                    Result error = null;

                    if (setup)
                    {
                        if (SaveTitle(ref error))
                        {
                            System.Console.Title = String.Format(
                                TitleFormat, RuntimeOps.IsAdministrator() ?
                                    "Administrator:" : String.Empty, DefaultTitle,
                                base.Title, RuntimeOps.GetCertificateSubject(
                                    GlobalState.GetAssembly(),
                                    RuntimeOps.ShouldCheckFileTrusted())).Trim();

                            return true;
                        }
                        else
                        {
                            DebugOps.Complain(interpreter, ReturnCode.Error, error);
                        }
                    }
                    else
                    {
                        if (RestoreTitle(ref error))
                            return true;
                        else
                            DebugOps.Complain(interpreter, ReturnCode.Error, error);
                    }
                }
                else
                {
                    return true; /* BUGFIX: Fake success. */
                }
            }
            catch
            {
                // do nothing.
            }

            return false;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Console Stream Handling
        #region Native Console Stream Handling
#if NATIVE && WINDOWS
        private bool IsChannelRedirected(
            ChannelType channelType
            )
        {
            if (PlatformOps.IsWindowsOperatingSystem())
            {
                Result error = null;
                IntPtr handle = ConsoleOps.GetHandle(channelType, ref error);

                if (NativeOps.IsValidHandle(handle))
                {
                    bool redirected = false;

                    if (ConsoleOps.IsHandleRedirected(handle,
                            ref redirected, ref error) == ReturnCode.Ok)
                    {
                        return redirected;
                    }
                }

                //
                // NOTE: Either we failed to get the handle or we could
                //       not determine if it has been redirected.
                //
                DebugOps.Complain(
                    SafeGetInterpreter(), ReturnCode.Error, error);
            }

            return false;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region System.Console Stream Handling
        private static Stream GetInputStream()
        {
            Stream stream = null;
            Result error = null;

            if (GetInputStream(ref stream, ref error) == ReturnCode.Ok)
                return stream;

            return null;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static ReturnCode GetInputStream(
            ref Stream stream,
            ref Result error
            )
        {
            //
            // HACK: Because the System.Console object in the .NET Framework provides
            //       no means to query the underlying input/output streams, we must
            //       do it here by force.
            //
            try
            {
                TextReader textReader = System.Console.In;

                if (textReader != null)
                {
                    StreamReader streamReader = textReader.GetType().InvokeMember(
                        CommonOps.Runtime.IsMono() ? "reader" : "_in",
                        MarshalOps.PrivateInstanceGetFieldBindingFlags,
                        null, textReader, null) as StreamReader;

                    if (streamReader != null)
                    {
                        stream = streamReader.BaseStream;

                        return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "failed to get internal console input stream reader";
                    }
                }
                else
                {
                    error = "failed to get console input text reader";
                }
            }
            catch (Exception e)
            {
                error = e;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static Stream GetOutputStream()
        {
            Stream stream = null;
            Result error = null;

            if (GetOutputStream(ref stream, ref error) == ReturnCode.Ok)
                return stream;

            return null;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static ReturnCode GetOutputStream(
            ref Stream stream,
            ref Result error
            )
        {
            //
            // HACK: Because the System.Console object in the .NET Framework provides
            //       no means to query the underlying input/output streams, we must
            //       do it here by force.
            //
            try
            {
                TextWriter textWriter = System.Console.Out;

                if (textWriter != null)
                {
                    StreamWriter streamWriter = textWriter.GetType().InvokeMember(
                        CommonOps.Runtime.IsMono() ? "writer" : "_out",
                        MarshalOps.PrivateInstanceGetFieldBindingFlags,
                        null, textWriter, null) as StreamWriter;

                    if (streamWriter != null)
                    {
                        stream = streamWriter.BaseStream;

                        return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "failed to get internal console output stream writer";
                    }
                }
                else
                {
                    error = "failed to get console output text writer";
                }
            }
            catch (Exception e)
            {
                error = e;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static Stream GetErrorStream()
        {
            Stream stream = null;
            Result error = null;

            if (GetErrorStream(ref stream, ref error) == ReturnCode.Ok)
                return stream;

            return null;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static ReturnCode GetErrorStream(
            ref Stream stream,
            ref Result error
            )
        {
            //
            // HACK: Because the System.Console object in the .NET Framework provides
            //       no means to query the underlying input/output streams, we must
            //       do it here by force.
            //
            try
            {
                TextWriter textWriter = System.Console.Error;

                if (textWriter != null)
                {
                    StreamWriter streamWriter = textWriter.GetType().InvokeMember(
                        CommonOps.Runtime.IsMono() ? "writer" : "_out",
                        MarshalOps.PrivateInstanceGetFieldBindingFlags,
                        null, textWriter, null) as StreamWriter;

                    if (streamWriter != null)
                    {
                        stream = streamWriter.BaseStream;

                        return ReturnCode.Ok;
                    }
                    else
                    {
                        error = "failed to get internal console error stream writer";
                    }
                }
                else
                {
                    error = "failed to get console error text writer";
                }
            }
            catch (Exception e)
            {
                error = e;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static ReturnCode ResetSystemConsole(
            ref Result error
            )
        {
            try
            {
                if (!CommonOps.Runtime.IsMono())
                {
                    //
                    // HACK: Force all the static handles used by the
                    //       System.Console class to be recreated.
                    //       There should be an official way of doing
                    //       this.
                    //
                    typeof(System.Console).InvokeMember("_consoleInputHandle",
                        MarshalOps.PrivateStaticSetFieldBindingFlags,
                        null, null, new object[] { IntPtr.Zero });

                    typeof(System.Console).InvokeMember("_consoleOutputHandle",
                        MarshalOps.PrivateStaticSetFieldBindingFlags,
                        null, null, new object[] { IntPtr.Zero });

                    typeof(System.Console).InvokeMember("_in",
                        MarshalOps.PrivateStaticSetFieldBindingFlags,
                        null, null, new object[] { null });

                    typeof(System.Console).InvokeMember("_out",
                        MarshalOps.PrivateStaticSetFieldBindingFlags,
                        null, null, new object[] { null });

                    typeof(System.Console).InvokeMember("_error",
                        MarshalOps.PrivateStaticSetFieldBindingFlags,
                        null, null, new object[] { null });

                    return ReturnCode.Ok;
                }
                else
                {
                    error = "not implemented";
                }
            }
            catch (Exception e)
            {
                error = e;
            }

            return ReturnCode.Error;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Static Stream "Factory" Methods
        private static StreamReader NewStreamReader(
            Stream stream,
            Encoding encoding
            )
        {
            if ((stream != null) && (encoding != null))
                return new StreamReader(stream, encoding);

            return null;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static StreamWriter NewStreamWriter(
            Stream stream,
            Encoding encoding
            )
        {
            if ((stream != null) && (encoding != null))
            {
                StreamWriter streamWriter = new StreamWriter(stream, encoding);

                streamWriter.AutoFlush = true;

                return streamWriter;
            }

            return null;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Console Setup Handling
        #region Console CancelKeyPress Handling
        #region Native Console CancelKeyPress Handling
#if NATIVE && WINDOWS
        private ReturnCode UnhookSystemConsoleControlHandler(
            ref Result error
            )
        {
            //
            // HACK: Because the System.Console object in the .NET Framework provides
            //       no means to unhook it from its native console callbacks, we must
            //       do it here by force.
            //
            try
            {
                if (!NoCancel)
                {
                    if (!CommonOps.Runtime.IsMono())
                    {
                        //
                        // NOTE: First, grab the private static ControlCHooker field from
                        //       the static System.Console object.
                        //
                        object hooker = typeof(System.Console).InvokeMember(
                            "_hooker", MarshalOps.PrivateStaticGetFieldBindingFlags,
                            null, null, null);

                        if (hooker != null)
                        {
                            //
                            // NOTE: Next, call the Unhook method of the ControlCHooker object
                            //       so that it will unhook itself from its native callbacks.
                            //
                            hooker.GetType().InvokeMember("Unhook",
                                MarshalOps.PrivateInstanceMethodBindingFlags,
                                null, hooker, null);

                            //
                            // NOTE: Finally, null out the private static ControlCHooker field
                            //       inside the System.Console object so that it will know when
                            //       it needs to be re-hooked later.
                            //
                            typeof(System.Console).InvokeMember("_hooker",
                                MarshalOps.PrivateStaticSetFieldBindingFlags,
                                null, null, new object[] { null });

                            return ReturnCode.Ok;
                        }
                        else
                        {
                            error = "failed to get internal console hook";
                        }
                    }
                    else
                    {
                        //
                        // TEST: This special handling should not be needed for Mono?
                        //
                        return ReturnCode.Ok; // NOTE: Fake success.
                    }
                }
                else
                {
                    return ReturnCode.Ok; // NOTE: Fake success.
                }
            }
            catch (Exception e)
            {
                error = e;
            }

            return ReturnCode.Error;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region System.Console CancelKeyPress Handling
        protected virtual bool SetupCancelKeyPressHandler(
            bool setup
            )
        {
            try
            {
                //
                // NOTE: Has setting up the script cancellation keypress been
                //       explicitly disabled?
                //
                if (!NoCancel)
                {
                    if (setup)
                        System.Console.CancelKeyPress +=
                            Interpreter.ConsoleCancelEventHandler;
                    else
                        System.Console.CancelKeyPress -=
                            Interpreter.ConsoleCancelEventHandler;

                    return true; // success.
                }
                else
                {
                    return true; // fake success.
                }
            }
            catch
            {
                return false; // failure.
            }
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Native Console Icon Handling
#if NATIVE && WINDOWS && DRAWING
        private bool SetupIcon()
        {
            try
            {
                //
                // NOTE: Has changing the icon been explicitly disabled?
                //
                if (!NoIcon)
                {
                    if (PlatformOps.IsWindowsOperatingSystem())
                    {
                        string packageName = GlobalState.GetPackageName(false);

                        if (!String.IsNullOrEmpty(packageName))
                        {
                            Stream stream = AssemblyOps.GetResourceStream(
                                GlobalState.GetAssembly(),
                                packageName + FileExtension.Icon);

                            return SetupIcon(true, stream);
                        }
                    }
                    else
                    {
                        return true; /* BUGFIX: Fake success. */
                    }
                }
                else
                {
                    return true; /* BUGFIX: Fake success. */
                }
            }
            catch
            {
                // do nothing.
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool SetupIcon(
            bool setup,
            Stream stream
            )
        {
            try
            {
                //
                // NOTE: Has changing the icon been explicitly disabled?
                //
                if (!NoIcon && IsOpen())
                {
                    if (PlatformOps.IsWindowsOperatingSystem())
                    {
                        if (setup)
                        {
                            if (stream != null)
                            {
                                if (icon != null)
                                {
                                    icon.Dispose();
                                    icon = null;
                                }

                                icon = new Icon(stream);

                                oldSmallIcon = WindowOps.UnsafeNativeMethods.SendMessage(
                                    GetConsoleWindow(),
                                    WindowOps.UnsafeNativeMethods.WM_GETICON,
                                    new UIntPtr(WindowOps.UnsafeNativeMethods.ICON_SMALL),
                                    IntPtr.Zero);

                                oldBigIcon = WindowOps.UnsafeNativeMethods.SendMessage(
                                    GetConsoleWindow(),
                                    WindowOps.UnsafeNativeMethods.WM_GETICON,
                                    new UIntPtr(WindowOps.UnsafeNativeMethods.ICON_BIG),
                                    IntPtr.Zero);

                                WindowOps.UnsafeNativeMethods.SendMessage(
                                    GetConsoleWindow(),
                                    WindowOps.UnsafeNativeMethods.WM_SETICON,
                                    new UIntPtr(WindowOps.UnsafeNativeMethods.ICON_SMALL),
                                    icon.Handle);

                                WindowOps.UnsafeNativeMethods.SendMessage(
                                    GetConsoleWindow(),
                                    WindowOps.UnsafeNativeMethods.WM_SETICON,
                                    new UIntPtr(WindowOps.UnsafeNativeMethods.ICON_BIG),
                                    icon.Handle);

                                return true;
                            }
                        }
                        else
                        {
                            WindowOps.UnsafeNativeMethods.SendMessage(
                                GetConsoleWindow(),
                                WindowOps.UnsafeNativeMethods.WM_SETICON,
                                new UIntPtr(WindowOps.UnsafeNativeMethods.ICON_BIG),
                                oldBigIcon);

                            oldBigIcon = IntPtr.Zero;

                            WindowOps.UnsafeNativeMethods.SendMessage(
                                GetConsoleWindow(),
                                WindowOps.UnsafeNativeMethods.WM_SETICON,
                                new UIntPtr(WindowOps.UnsafeNativeMethods.ICON_SMALL),
                                oldSmallIcon);

                            oldSmallIcon = IntPtr.Zero;

                            if (icon != null)
                            {
                                icon.Dispose();
                                icon = null;
                            }

                            return true;
                        }
                    }
                    else
                    {
                        return true; /* NOTE: Fake success. */
                    }
                }
                else
                {
                    return true; /* BUGFIX: Fake success. */
                }
            }
            catch
            {
                // do nothing.
            }

            return false;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Native Console Mode Handling
#if NATIVE && WINDOWS
        protected bool ChangeMode(
            ChannelType channelType,
            bool enable,
            uint mode
            )
        {
            CheckDisposed();

            if (IsOpen() &&
                PlatformOps.IsWindowsOperatingSystem() &&
                (ConsoleOps.ChangeMode(channelType, enable, mode) == ReturnCode.Ok))
            {
                return true;
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private bool SetupMode()
        {
            if (IsOpen())
            {
                if (PlatformOps.IsWindowsOperatingSystem())
                {
                    //
                    // HACK: Disable this flag so that right-click works properly in
                    //       the shell (i.e. it brings up the context menu, just like
                    //       it does by default in cmd.exe).
                    //
                    uint mode = ConsoleOps.UnsafeNativeMethods.ENABLE_MOUSE_INPUT;

                    if (ConsoleOps.ChangeMode(ChannelType.Input, false, mode) != ReturnCode.Ok)
                        return false;
                }
            }

            return true; // NOTE: Fake success.
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private static IntPtr GetConsoleWindow()
        {
            try
            {
                if (PlatformOps.IsWindowsOperatingSystem())
                    return ConsoleOps.UnsafeNativeMethods.GetConsoleWindow();
            }
            catch
            {
                // do nothing.
            }

            return IntPtr.Zero;
        }
#endif
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Master Setup Methods
        private static bool Setup(
            Console host,
            bool setup,
            bool force
            )
        {
            if (setup)
            {
                if (host != null)
                {
                    int newReferenceCount =
                        Interlocked.Increment(ref referenceCount);

                    if (force || (newReferenceCount == 1))
                        return host.SetupTitle(true) &&
#if NATIVE && WINDOWS && DRAWING
                            host.SetupIcon() &&
#endif
#if NATIVE && WINDOWS
                            host.SetupMode() &&
#endif
                            host.SetupCancelKeyPressHandler(true);
                    else
                        return true;
                }
            }
            else
            {
                if (host != null)
                {
                    int newReferenceCount =
                        Interlocked.Decrement(ref referenceCount);

                    if (force || (newReferenceCount <= 0))
                        return host.SetupCancelKeyPressHandler(false) &&
#if NATIVE && WINDOWS && DRAWING
                            host.SetupIcon(false, null) &&
#endif
                            host.SetupTitle(false);
                    else
                        return true;
                }
            }

            return false;
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region Console Test Mode Handling
        internal void EnableTests(
            bool enable
            )
        {
            hostFlags = GetFlags();
            hostTestFlags = GetTestFlags();

            if (enable)
            {
                //
                // NOTE: Enable test mode.
                //
                hostFlags |= HostFlags.Test;

                //
                // NOTE: Enable each of the individual tests.
                //
                hostTestFlags |= HostTestFlags.TestCustomInfo;
            }
            else
            {
                //
                // NOTE: Disable test mode.
                //
                hostFlags &= ~HostFlags.Test;

                //
                // NOTE: Disable each of the individual tests.
                //
                hostTestFlags &= ~HostTestFlags.TestCustomInfo;
            }
        }
        #endregion
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IInteractiveHost Members
        public override string Title
        {
            set
            {
                CheckDisposed();

                base.Title = value;
                SetupTitle(true);
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool IsInputRedirected()
        {
            CheckDisposed();

            try
            {
#if MONO_BUILD
#pragma warning disable 219
#endif
                /* IGNORED */
                bool keyAvailable = System.Console.KeyAvailable; // NOTE: Flagged by the Mono C# compiler.
#if MONO_BUILD
#pragma warning restore 219
#endif

                //
                // NOTE: If we got this far, input has not been
                //       redirected (i.e. there was no exception
                //       thrown by KeyAvailable).
                //
                return false;
            }
            catch (InvalidOperationException)
            {
                //
                // NOTE: Per MSDN, input is being redirected from
                //       a "file".
                //
                return true;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool IsOpen()
        {
            CheckDisposed();

#if NATIVE && WINDOWS
            if (PlatformOps.IsWindowsOperatingSystem())
                return GetConsoleWindow() != IntPtr.Zero;
            else
                return true;
#else
            return true;
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool Pause()
        {
            CheckDisposed();

            try
            {
                System.Console.ReadKey(true);

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private HostFlags hostFlags = HostFlags.Invalid;
        public override HostFlags GetFlags()
        {
            CheckDisposed();

            if (hostFlags == HostFlags.Invalid)
            {
                //
                // NOTE: We support text output, colors, positioning,
                //       sizing, and resizing.
                //
                hostFlags = HostFlags.Resizable | HostFlags.Color |
                            HostFlags.Text | HostFlags.Sizing |
                            HostFlags.Positioning | base.GetFlags();

                if ((WindowWidth >= SuperJumboWindowWidth) &&
                    (WindowHeight >= SuperJumboWindowHeight))
                {
                    hostFlags |= HostFlags.SuperJumboSize;
                }
                else if ((WindowWidth >= JumboWindowWidth) &&
                    (WindowHeight >= JumboWindowHeight))
                {
                    hostFlags |= HostFlags.JumboSize;
                }
                else if ((WindowWidth >= SuperFullWindowWidth) &&
                    (WindowHeight >= SuperFullWindowHeight))
                {
                    hostFlags |= HostFlags.SuperFullSize;
                }
                else if ((WindowWidth >= FullWindowWidth) &&
                    (WindowHeight >= FullWindowHeight))
                {
                    hostFlags |= HostFlags.FullSize;
                }
                else if ((WindowWidth >= CompactWindowWidth) &&
                    (WindowHeight >= CompactWindowHeight))
                {
                    hostFlags |= HostFlags.CompactSize;
                }
                else if ((WindowWidth >= MinimumWindowWidth) &&
                    (WindowHeight >= MinimumWindowHeight))
                {
                    hostFlags |= HostFlags.MinimumSize;
                }
                else if (!PlatformOps.IsWindowsOperatingSystem())
                {
                    //
                    // NOTE: No idea on this platform as Mono does not support
                    //       console window width and height on Unix (?).  Fake
                    //       it.
                    //
                    hostFlags |= HostFlags.CompactSize;
                }
                else
                {
                    //
                    // NOTE: We should not get here.
                    //
                    hostFlags |= HostFlags.ZeroSize;
                }
            }

            //
            // WARNING: Do not use the InTestMode method here, it calls
            //          this method.
            //
            if (FlagOps.HasFlags(hostFlags, HostFlags.Test, true))
                hostFlags |= HostFlags.CustomInfo;
            else
                hostFlags &= ~HostFlags.CustomInfo;

            return hostFlags;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool ReadLine(
            ref string value
            )
        {
            CheckDisposed();

            try
            {
                value = System.Console.ReadLine();

                if (Echo)
                    System.Console.WriteLine(value);

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool WriteLine()
        {
            CheckDisposed();

            try
            {
                System.Console.WriteLine();

                return true;
            }
            catch
            {
                return false;
            }
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IStreamHost Members
        public override Stream DefaultIn
        {
            get { CheckDisposed(); return System.Console.OpenStandardInput(); }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override Stream DefaultOut
        {
            get { CheckDisposed(); return System.Console.OpenStandardOutput(); }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override Stream DefaultError
        {
            get { CheckDisposed(); return System.Console.OpenStandardError(); }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override Stream In
        {
            get { CheckDisposed(); return GetInputStream(); }
            set { CheckDisposed(); System.Console.SetIn(NewStreamReader(value, InputEncoding)); }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override Stream Out
        {
            get { CheckDisposed(); return GetOutputStream(); }
            set { CheckDisposed(); System.Console.SetOut(NewStreamWriter(value, OutputEncoding)); }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override Stream Error
        {
            get { CheckDisposed(); return GetErrorStream(); }
            set { CheckDisposed(); System.Console.SetError(NewStreamWriter(value, ErrorEncoding)); }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override Encoding InputEncoding
        {
            get { CheckDisposed(); return System.Console.InputEncoding; }
            set { CheckDisposed(); System.Console.InputEncoding = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override Encoding OutputEncoding
        {
            get { CheckDisposed(); return System.Console.OutputEncoding; }
            set { CheckDisposed(); System.Console.OutputEncoding = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override Encoding ErrorEncoding
        {
            get { CheckDisposed(); return System.Console.OutputEncoding; }
            set { CheckDisposed(); System.Console.OutputEncoding = value; }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool ResetIn()
        {
            CheckDisposed();

            try
            {
                Encoding encoding = System.Console.InputEncoding;
                System.Console.InputEncoding = encoding;

#if MONO_BUILD
#pragma warning disable 219
#endif
                /* IGNORED */
                TextReader textReader = System.Console.In; // NOTE: Flagged by the Mono C# compiler.
#if MONO_BUILD
#pragma warning restore 219
#endif

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool ResetOut()
        {
            CheckDisposed();

            try
            {
                Encoding encoding = System.Console.OutputEncoding;
                System.Console.OutputEncoding = encoding;

#if MONO_BUILD
#pragma warning disable 219
#endif
                /* IGNORED */
                TextWriter textWriter = System.Console.Out; // NOTE: Flagged by the Mono C# compiler.
#if MONO_BUILD
#pragma warning restore 219
#endif

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool ResetError()
        {
            CheckDisposed();

            try
            {
                Encoding encoding = System.Console.OutputEncoding;
                System.Console.OutputEncoding = encoding;

#if MONO_BUILD
#pragma warning disable 219
#endif
                /* IGNORED */
                TextWriter textWriter = System.Console.Error; // NOTE: Flagged by the Mono C# compiler.
#if MONO_BUILD
#pragma warning restore 219
#endif

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool IsOutputRedirected()
        {
            CheckDisposed();

#if NATIVE && WINDOWS
            return IsChannelRedirected(ChannelType.Output);
#else
            return false; /* NOT YET IMPLEMENTED */
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool IsErrorRedirected()
        {
            CheckDisposed();

#if NATIVE && WINDOWS
            return IsChannelRedirected(ChannelType.Error);
#else
            return false; /* NOT YET IMPLEMENTED */
#endif
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IDebugHost Members
        public override IHost Clone(
            Interpreter interpreter
            )
        {
            CheckDisposed();

            return new Console(new HostData(
                Name, Group, Description, ClientData, typeof(Console).Name, interpreter,
                ResourceManager, Profile, NoColor, NoTitle, NoIcon, NoProfile, NoCancel));
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        private HostTestFlags hostTestFlags = HostTestFlags.Invalid;
        public override HostTestFlags GetTestFlags()
        {
            CheckDisposed();

            if (hostTestFlags == HostTestFlags.Invalid)
                hostTestFlags = HostTestFlags.None;

            return hostTestFlags;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override ReturnCode Cancel(
            ref Result error
            )
        {
            CheckDisposed();

#if NATIVE && WINDOWS
            if (PlatformOps.IsWindowsOperatingSystem())
            {
                ReturnCode code = ConsoleOps.SendControlEvent(
                    ControlEvent.CTRL_C_EVENT, ref error);

                if (code == ReturnCode.Ok)
                {
                    IntPtr handle = GetConsoleWindow();

                    if (handle != IntPtr.Zero)
                    {
                        //
                        // NOTE: This is an attempt to "nicely" break out of the
                        //       synchronous Console.ReadLine call so that the
                        //       interactive loop can realize any changes in the
                        //       interpreter state (i.e. has the interpreter been
                        //       marked as "exited"?).
                        //
                        code = WindowOps.SimulateReturnKey(handle, ref error);
                    }
                    else
                    {
                        error = "invalid console window";
                        code = ReturnCode.Error;
                    }
                }

                return code;
            }
            else
            {
                error = "not implemented";
                return ReturnCode.Error;
            }
#else
            error = "not implemented";
            return ReturnCode.Error;
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override ReturnCode Exit(
            ref Result error
            )
        {
            CheckDisposed();

#if NATIVE && WINDOWS
            if (PlatformOps.IsWindowsOperatingSystem())
            {
                try
                {
                    Interpreter interpreter = SafeGetInterpreter();

                    if (interpreter != null)
                    {
                        //
                        // NOTE: Stop any further activity in the interpreter.
                        //
                        interpreter.Exit = true;

                        //
                        // NOTE: Bail out of Console.ReadLine, etc.
                        //
                        return ConsoleOps.CloseStandardInput(ref error);
                    }
                    else
                    {
                        error = "invalid interpreter";
                    }
                }
                catch (Exception e)
                {
                    error = e;
                }

                return ReturnCode.Error;
            }
            else
            {
                error = "not implemented";
                return ReturnCode.Error;
            }
#else
            error = "not implemented";
            return ReturnCode.Error;
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool WriteDebugLine()
        {
            CheckDisposed();

            //
            // TODO: We have no dedicated place for debug output;
            //       therefore, just forward it as normal output.
            //
            return WriteLine();
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool WriteDebug(
            char value,
            bool newLine
            )
        {
            CheckDisposed();

            //
            // TODO: We have no dedicated place for debug output;
            //       therefore, just forward it as normal output
            //       [with the correct colors].
            //
            return Write(value, 1, newLine, DebugForegroundColor, DebugBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool WriteDebug(
            string value,
            bool newLine
            )
        {
            CheckDisposed();

            //
            // TODO: We have no dedicated place for debug output;
            //       therefore, just forward it as normal output
            //       [with the correct colors].
            //
            return Write(value, newLine, DebugForegroundColor, DebugBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool WriteErrorLine()
        {
            CheckDisposed();

            try
            {
                System.Console.Error.WriteLine();

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool WriteError(
            char value,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteCore(
                System.Console.Error.Write, System.Console.Error.WriteLine,
                value, 1, newLine, ErrorForegroundColor, ErrorBackgroundColor);
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool WriteError(
            string value,
            bool newLine
            )
        {
            CheckDisposed();

            return WriteCore(
                System.Console.Error.Write, System.Console.Error.WriteLine,
                value, newLine, ErrorForegroundColor, ErrorBackgroundColor);
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IInformationHost Members
        public override bool WriteCustomInfo(
            Interpreter interpreter,
            bool empty,
            bool newLine,
            ConsoleColor foregroundColor,
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

#if TEST
            if (InTestMode() && HasTestFlags(HostTestFlags.TestCustomInfo, true))
                return _Tests.Default.TestWriteCustomInfo(interpreter, empty, newLine,
                    foregroundColor, backgroundColor);
#endif

            return true;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IHost Members
        public override bool Beep(
            int frequency,
            int duration
            )
        {
            CheckDisposed();

            try
            {
                System.Console.Beep(frequency, duration);

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool IsIdle()
        {
            CheckDisposed();

            //
            // STUB: We have no better idle detection.
            //
            return true;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool Clear()
        {
            CheckDisposed();

            try
            {
                System.Console.Clear();

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool ResetFlags()
        {
            CheckDisposed();

            hostFlags = HostFlags.Invalid;

            return base.ResetFlags();
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool GetPosition(
            ref int left,
            ref int top
            )
        {
            CheckDisposed();

            try
            {
                left = System.Console.CursorLeft;
                top = System.Console.CursorTop;

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool SetPosition(
            int left,
            int top
            )
        {
            CheckDisposed();

            try
            {
                if ((left != _Position.Invalid) && (top != _Position.Invalid))
                    System.Console.SetCursorPosition(left, top);
                else if (left != _Position.Invalid)
                    System.Console.CursorLeft = left;
                else if (top != _Position.Invalid)
                    System.Console.CursorTop = top;

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool ResetSize()
        {
            CheckDisposed();

            try
            {
                //
                // NOTE: Make sure we successfully saved the original buffer
                //       and window sizes earlier.
                //
                if ((savedBufferWidth != _Size.Invalid) &&
                    (savedBufferHeight != _Size.Invalid) &&
                    (savedWindowWidth != _Size.Invalid) &&
                    (savedWindowHeight != _Size.Invalid))
                {
                    //
                    // NOTE: Set the window size to the minimum possible so
                    //       that any buffer size we set (within bounds) is
                    //       valid.
                    //
                    System.Console.SetWindowSize(1, 1);

                    //
                    // NOTE: Restore the original buffer size.
                    //
                    System.Console.SetBufferSize(savedBufferWidth, savedBufferHeight);

                    //
                    // NOTE: Restore the original window size.
                    //
                    System.Console.SetWindowSize(savedWindowWidth, savedWindowHeight);

                    //
                    // NOTE: If we got this far, we succeeded.
                    //
                    return true;
                }
            }
            catch
            {
                // do nothing.
            }

            return false;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool GetSize(
            ref int width,
            ref int height
            )
        {
            CheckDisposed();

            try
            {
                width = System.Console.BufferWidth;
                height = System.Console.BufferHeight;

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool SetSize(
            int width,
            int height
            )
        {
            CheckDisposed();

            try
            {
                int newWindowWidth = width;
                int newWindowHeight = height;

                //
                // NOTE: If the caller does not want to set the width (i.e.
                //       it is invalid) then use the current window width.
                //
                if (newWindowWidth == _Size.Invalid)
                    newWindowWidth = System.Console.WindowWidth;

                //
                // NOTE: If the caller does not want to set the height (i.e.
                //       it is invalid) then use the current window height.
                //
                if (newWindowHeight == _Size.Invalid)
                    newWindowHeight = System.Console.WindowHeight;

                int newBufferWidth = width;
                int newBufferHeight = height;

                //
                // NOTE: If the caller does not want to set the width (i.e.
                //       it is invalid) then use the current buffer width.
                //
                if (newBufferWidth == _Size.Invalid)
                    newBufferWidth = System.Console.BufferWidth;

                //
                // NOTE: If the caller does not want to set the height (i.e.
                //       it is invalid) then use the current buffer height.
                //
                if (newBufferHeight == _Size.Invalid)
                    newBufferHeight = System.Console.BufferHeight;

                //
                // NOTE: Set the window size to the minimum possible so
                //       that any buffer size we set (within bounds) is
                //       valid.
                //
                System.Console.SetWindowSize(1, 1);

                //
                // NOTE: Set the new buffer size.
                //
                System.Console.SetBufferSize(newBufferWidth, newBufferHeight);

                //
                // NOTE: Try to make sure that we do not attempt to set an
                //       unreasonable window width.
                //
                if ((newWindowWidth > System.Console.BufferWidth) ||
                    (newWindowWidth > System.Console.LargestWindowWidth))
                {
                    newWindowWidth = Math.Min(
                        System.Console.BufferWidth,
                        System.Console.LargestWindowWidth);
                }

                //
                // NOTE: Try to make sure that we do not attempt to set an
                //       unreasonable window height.
                //
                if ((newWindowHeight > System.Console.BufferHeight) ||
                    (newWindowHeight > System.Console.LargestWindowHeight))
                {
                    newWindowHeight = Math.Min(
                        System.Console.BufferHeight,
                        System.Console.LargestWindowHeight);
                }

                //
                // NOTE: Set the new window size.
                //
                System.Console.SetWindowSize(newWindowWidth, newWindowHeight);

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override ReturnCode GetMode(
            ChannelType channelType,
            ref uint mode,
            ref Result error
            )
        {
            CheckDisposed();

#if NATIVE && WINDOWS
            if (PlatformOps.IsWindowsOperatingSystem())
            {
                return ConsoleOps.GetMode(channelType, ref mode, ref error);
            }
            else
            {
                error = "not implemented";
                return ReturnCode.Error;
            }
#else
            error = "not implemented";
            return ReturnCode.Error;
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override ReturnCode SetMode(
            ChannelType channelType,
            uint mode,
            ref Result error
            )
        {
            CheckDisposed();

#if NATIVE && WINDOWS
            if (PlatformOps.IsWindowsOperatingSystem())
            {
                return ConsoleOps.SetMode(channelType, mode, ref error);
            }
            else
            {
                error = "not implemented";
                return ReturnCode.Error;
            }
#else
            error = "not implemented";
            return ReturnCode.Error;
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool ResetColors()
        {
            CheckDisposed();

            try
            {
                System.Console.ResetColor();

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool GetColors(
            ref ConsoleColor foregroundColor,
            ref ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            try
            {
                foregroundColor = System.Console.ForegroundColor;
                backgroundColor = System.Console.BackgroundColor;

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool SetForegroundColor(
            ConsoleColor foregroundColor
            )
        {
            CheckDisposed();

            try
            {
                if (foregroundColor != _ConsoleColor.None)
                    System.Console.ForegroundColor = foregroundColor;

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool SetBackgroundColor(
            ConsoleColor backgroundColor
            )
        {
            CheckDisposed();

            try
            {
                if (backgroundColor != _ConsoleColor.None)
                    System.Console.BackgroundColor = backgroundColor;

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override ReturnCode Open(
            ref Result error
            )
        {
            CheckDisposed();

#if NATIVE && WINDOWS
            if (PlatformOps.IsWindowsOperatingSystem())
            {
                ReturnCode code = ConsoleOps.Open(ref error);

                if (code == ReturnCode.Ok)
                {
                    //
                    // NOTE: Now, re-setup our console customizations.
                    //
                    if (!Setup(this, true, true))
                    {
                        error = "failed to re-setup console";
                        code = ReturnCode.Error;
                    }
                }

                return code;
            }
            else
            {
                error = "not implemented";
                return ReturnCode.Error;
            }
#else
            error = "not implemented";
            return ReturnCode.Error;
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override ReturnCode Close(
            ref Result error
            )
        {
            CheckDisposed();

#if NATIVE && WINDOWS
            if (PlatformOps.IsWindowsOperatingSystem())
            {
                ReturnCode code;

                if (Setup(this, false, true))
                {
                    code = UnhookSystemConsoleControlHandler(ref error);

                    if (code == ReturnCode.Ok)
                        code = ConsoleOps.Close(ref error);
                }
                else
                {
                    error = "failed to un-setup console";
                    code = ReturnCode.Error;
                }

                return code;
            }
            else
            {
                error = "not implemented";
                return ReturnCode.Error;
            }
#else
            error = "not implemented";
            return ReturnCode.Error;
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override ReturnCode Discard(
            ref Result error
            )
        {
            CheckDisposed();

#if NATIVE && WINDOWS
            if (PlatformOps.IsWindowsOperatingSystem())
            {
                if (!CommonOps.Runtime.IsMono())
                {
                    try
                    {
                        object cachedInputRecord = typeof(System.Console).InvokeMember(
                           "_cachedInputRecord", MarshalOps.PrivateStaticGetFieldBindingFlags,
                           null, null, null);

                        if (cachedInputRecord != null)
                        {
                            Marshal.WriteInt16(cachedInputRecord, 0, 0);
                        }
                        else
                        {
                            error = "failed to get internal cached input record";
                            return ReturnCode.Error;
                        }
                    }
                    catch (Exception e)
                    {
                        error = e;
                        return ReturnCode.Error;
                    }
                }

                return ConsoleOps.FlushInputBuffer(ref error);
            }
            else
            {
                error = "not implemented";
                return ReturnCode.Error;
            }
#else
            error = "not implemented";
            return ReturnCode.Error;
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override ReturnCode Reset(
            ref Result error
            )
        {
            CheckDisposed();

            if ((ResetSystemConsole(ref error) == ReturnCode.Ok) &&
                (base.Reset(ref error) == ReturnCode.Ok))
            {
                hostFlags = HostFlags.Invalid;
                return ReturnCode.Ok;
            }

            return ReturnCode.Error;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool Read(
            ref int value
            )
        {
            CheckDisposed();

            try
            {
                value = System.Console.Read();

                if (Echo)
                    System.Console.Write(Convert.ToChar(value));

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool ReadKey(
            bool intercept,
            ref IClientData clientData
            )
        {
            CheckDisposed();

            try
            {
                clientData = new ClientData(System.Console.ReadKey(intercept)); /* throw */

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        [Obsolete()]
        public override bool ReadKey(
            bool intercept,
            ref ConsoleKeyInfo keyInfo
            )
        {
            CheckDisposed();

            try
            {
                keyInfo = System.Console.ReadKey(intercept);

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool Write(
            char value,
            bool newLine
            )
        {
            CheckDisposed();

            try
            {
                if (newLine)
                    System.Console.WriteLine(value);
                else
                    System.Console.Write(value);

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool Write(
            string value,
            bool newLine
            )
        {
            CheckDisposed();

            try
            {
                if (newLine)
                    System.Console.WriteLine(value);
                else
                    System.Console.Write(value);

                return true;
            }
            catch
            {
                return false;
            }
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool BeginSection(
            string name,
            IClientData clientData
            )
        {
            CheckDisposed();

            return true;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool EndSection(
            string name,
            IClientData clientData
            )
        {
            CheckDisposed();

            return true;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool BeginBox(
            string name,
            StringPairList list,
            IClientData clientData
            )
        {
            CheckDisposed();

            return true;
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        public override bool EndBox(
            string name,
            StringPairList list,
            IClientData clientData
            )
        {
            CheckDisposed();

            return true;
        }
        #endregion

        ///////////////////////////////////////////////////////////////////////////////////////////////

        #region IDisposable "Pattern" Members
        private bool disposed;
        private void CheckDisposed() /* throw */
        {
#if THROW_ON_DISPOSED
            if (disposed && Engine.IsThrowOnDisposed(SafeGetInterpreter(), true))
                throw new ObjectDisposedException(typeof(Console).Name);
#endif
        }

        ///////////////////////////////////////////////////////////////////////////////////////////////

        protected override void Dispose(bool disposing)
        {
            try
            {
                if (!disposed)
                {
                    //if (disposing)
                    //{
                    //    ////////////////////////////////////
                    //    // dispose managed resources here...
                    //    ////////////////////////////////////
                    //}

                    //////////////////////////////////////
                    // release unmanaged resources here...
                    //////////////////////////////////////

                    Setup(this, false, false);

                    disposed = true;
                }
            }
            finally
            {
                base.Dispose(disposing);
            }
        }
        #endregion
    }
}
